local MAJOR_VERSION = "LibSidePanel-1.0"
local MINOR_VERSION = tonumber(("$Revision: 1 $"):match("(%d+)")) + 0

if not LibStub then error(MAJOR_VERSION .. " requires LibStub") end

local oldLib = LibStub:GetLibrary(MAJOR_VERSION, true)
-- Try to create a new library
local sidePanel = LibStub:NewLibrary(MAJOR_VERSION, MINOR_VERSION)
-- Abort if we are no new library
if not sidePanel then
	return
end

local jostle = LibStub:GetLibrary("LibJostle-3.0", true)
local _G = _G

if LibStub('LibDebugLog-1.0') then
	LibStub('LibDebugLog-1.0'):Embed(sidePanel)
	sidePanel:ToggleDebugLog(true)
else
	function sidePanel:Debug (...)
	end
end

-- Preserve our old managed frames or create a new array
-- the content is: { frame, side, priority }
sidePanel.sideFrames = oldLib and oldLib.sideFrames or {}

-- Preserve our old update frame
sidePanel.frame = oldLib and oldLib.frame or CreateFrame("Frame")
local updateFrame = sidePanel.frame

-- Copied from LibJostle-3.0:
-- Start time, used for init stuff
local start = GetTime()
local nextTime = 0

-- Stores whether we have completed our initialization
local fullyInitted = false
updateFrame:SetScript("OnUpdate", function(this, elapsed)
	local now = GetTime()
	if now - start > 0 then
		fullyInitted = true
		-- set the schedule script
		this:SetScript("OnUpdate", function(this, elapsed)
			if GetTime() >= nextTime then
				this:Hide()
				sidePanel:Refresh()
			end
		end)
	end
end)

-- Shedules an update
function updateFrame:Schedule(time)
	time = time or 0
	nextTime = GetTime() + time
	self:Show()
end

-- Unregister all events, not sure why though
updateFrame:UnregisterAllEvents()

local function findFrame(frame)
	for k,v in ipairs(sidePanel.sideFrames) do
		if v.frame and v.frame == frame then
			return v
		end
	end
	return nil
end


--- Registers a frame.
-- @param frame the frame to register
-- @param side the side, must bei one of 'TOP','LEFT','BOTTOM', or 'RIGHT'
-- @param priority lower priority frames are closer to the edges of the screen (default is 100), must not be negative
-- @usage adds a frame to be managed, not that the frame must set its height or width (depending on the side) but should not set any other size property
--
function sidePanel:Register(frame, side, priority)
	-- check for valid side
	if side ~= 'TOP' and side ~= 'LEFT' and side ~= 'BOTTOM' and side ~= 'RIGHT' then
		error("Not a valid side: " .. side)
	end
	
	if not frame then
		error("Not a valid frame")
	end
	
	if priority < 0 then
		error("Priority must not be negative")
	end		
	
	local entry = findFrame(frame)
	local new = false
	if not entry then
		entry = {}
		new = true
	end
	entry.frame = frame
	entry.side = side
	entry.priority = priority or 100
	if new then
		table.insert(self.sideFrames, entry)
	end
	-- Add to jostle if present
	if jostle then
		if side == 'TOP' then
			jostle:RegisterTop(frame)
		elseif side == 'BOTTOM' then
			jostle:RegisterBottom(frame)
		end
	end
	
-- 	self:Debug("After insert:")
-- 	for k,entry in ipairs(self.sideFrames) do
-- 		self:Debug(k,entry.frame,entry.side,entry.priority)
-- 	end
	
	updateFrame:Schedule()
	return
end

--- Unregisters a frame
-- Well also remove the frame from jostle
-- @param frame the frame to unregister
function sidePanel:Unregister(frame)
	for k,v in ipairs(self.sideFrames) do
		if v.frame and v.frame == frame then
			table.remove(self.sideFrames, k)
			break
		end
	end
	-- Remove from jostle if present
	if jostle then
		jostle:Unregister(frame)
	end
	updateFrameSchedule()
end

local function sortFunction(a,b)
	return a.priority < b.priority
end

--- Will re anchor all managed frames. 
-- Should not be called directly, as it will be called by Register and Unregister.
function sidePanel:Refresh()
	-- Sort our frames
-- 	self:Debug("Before sorting:")
-- 	for k,entry in ipairs(self.sideFrames) do
-- 		self:Debug(k,entry.frame,entry.side,entry.priority)
-- 	end
-- 	table.sort(self.sideFrames,sortFunction)
-- 	self:Debug("After sorting:")
-- 	for k,entry in ipairs(self.sideFrames) do
-- 		self:Debug(k,entry.frame,entry.side,entry.priority)
-- 	end
	
	-- Get UIParent as start border:
	local margin = {}
	margin.tl = { frame = UIParent, point = 'TOPLEFT'     }
	margin.tr = { frame = UIParent, point = 'TOPRIGHT'    }
	margin.bl = { frame = UIParent, point = 'BOTTOMLEFT'  }
	margin.br = { frame = UIParent, point = 'BOTTOMRIGHT' }
	
	-- Iterate over each entry
	for k,entry in ipairs(self.sideFrames) do
		local side  = entry.side
		local frame = entry.frame
		-- Depending on the side, we need to update our margins
		frame:ClearAllPoints()
		if side == 'TOP' then
			frame:SetPoint( 'TOPLEFT',    margin.tl.frame, margin.tl.point, 0, 0 )
			frame:SetPoint( 'RIGHT',      margin.tr.frame, margin.tr.point, 0, 0 )
			margin.tl.frame = frame
			margin.tl.point = 'BOTTOMLEFT'
			margin.tr.frame = frame
			margin.tr.point = 'BOTTOMRIGHT'
		elseif side == 'LEFT' then
			frame:SetPoint( 'TOPLEFT',    margin.tl.frame, margin.tl.point, 0, 0 )
			frame:SetPoint( 'BOTTOM',     margin.bl.frame, margin.bl.point, 0, 0 )
			margin.tl.frame = frame
			margin.tl.point = 'TOPRIGHT'
			margin.bl.frame = frame
			margin.bl.point = 'BOTTOMRIGHT'	
		elseif side == 'BOTTOM' then
			frame:SetPoint( 'BOTTOMLEFT', margin.bl.frame, margin.bl.point, 0, 0 )
			frame:SetPoint( 'RIGHT',      margin.br.frame, margin.br.point, 0, 0 )
			margin.bl.frame = frame
			margin.bl.point = 'TOPLEFT'
			margin.br.frame = frame
			margin.br.point = 'TOPRIGHT'	
		elseif side == 'RIGHT' then
			frame:SetPoint( 'TOPRIGHT',   margin.tr.frame, margin.tr.point, 0, 0 )
			frame:SetPoint( 'BOTTOM',     margin.br.frame, margin.br.point, 0, 0 )
			margin.tr.frame = frame
			margin.tr.point = 'TOPLEFT'
			margin.br.frame = frame
			margin.br.point = 'BOTTOMLEFT'	
		else 
			error("Unexpected side")
		end
	end
end

--[==[
function sidePanel:GetScreenTop()
	local bottom = GetScreenHeight()
	for _,frame in ipairs(self.topFrames) do
		if frame.IsShown and frame:IsShown() and frame.GetBottom and frame:GetBottom() and frame:GetBottom() < bottom then
			bottom = frame:GetBottom()
		end
	end
	return bottom
end

function Jostle:GetScreenBottom()
	local top = 0
	for _,frame in ipairs(self.bottomFrames) do
		if frame.IsShown and frame:IsShown() and frame.GetTop and frame:GetTop() and frame:GetTop() > top then
			top = frame:GetTop()
		end
	end
	return top
end



function Jostle:IsTopAdjusting()
	return self.topAdjust
end

function Jostle:EnableTopAdjusting()
	if not self.topAdjust then
		self.topAdjust = not self.topAdjust
		JostleFrame:Schedule()
	end
end

function Jostle:DisableTopAdjusting()
	if self.topAdjust then
		self.topAdjust = not self.topAdjust
		JostleFrame:Schedule()
	end
end

function Jostle:IsBottomAdjusting()
	return self.bottomAdjust
end

function Jostle:EnableBottomAdjusting()
	if not self.bottomAdjust then
		self.bottomAdjust = not self.bottomAdjust
		JostleFrame:Schedule()
	end
end

function Jostle:DisableBottomAdjusting()
	if self.bottomAdjust then
		self.bottomAdjust = not self.bottomAdjust
		JostleFrame:Schedule()
	end
end

local tmp = {}
local queue = {}
local inCombat = false
function Jostle:ProcessQueue()
	if not inCombat and HasFullControl() then
		for k in pairs(queue) do
			self:Refresh(k)
			queue[k] = nil
		end
	end
end
function Jostle:PLAYER_CONTROL_GAINED()
	self:ProcessQueue()
end

function Jostle:PLAYER_REGEN_ENABLED()
	inCombat = false
	self:ProcessQueue()
end

function Jostle:PLAYER_REGEN_DISABLED()
	inCombat = true
end

local function isClose(alpha, bravo)
	return math.abs(alpha - bravo) < 0.1
end
]==]

